@@ -233,6 +233,15 @@ EAI_FAIL = getattr(socket, 'EAI_FAIL', NO_OBJ)
 # Constant for POPSSL
 POP3_SSL_PORT = 995
 
+
+# Python added poplib._MAXLINE somewhere along the way.  As far as I can
+# see, it serves no purpose except to introduce bugs into any software
+# using poplib.  Any computer running Python will have at least some megabytes
+# of userspace memory; arbitrarily causing message retrieval to break if any
+# "line" exceeds 2048 bytes is absolutely stupid.
+poplib._MAXLINE = 1 << 20   # 1MB; decrease this if you're running on a VIC-20
+
+
 #
 # Mix-in classes
 #
@@ -390,8 +399,11 @@ class Py24POP3SSLinitMixIn(object):
 
             # Ensure cert is for server we're connecting to
             if ssl and self.conf['ca_certs']:
-                ssl_match_hostname(self.conn.sslobj.getpeercert(),
-                               self.conf['server'])
+                ssl_match_hostname(
+                    self.conn.sslobj.getpeercert(),
+                    self.conf.get('ssl_cert_hostname', None) 
+                        or self.conf['server']
+                )
 
             if ssl_fingerprints:
                 if not actual_hash:
@@ -421,8 +433,19 @@ class Py24POP3SSLinitMixIn(object):
             )
 
         self.conn.sock.setblocking(1)
-        self.log.trace('POP3 connection %s established' % self.conn
-                       + os.linesep)
+
+        fingerprint_message = ('POP3 SSL connection %s established'
+                               % self.conn)
+        if actual_hash:
+            fingerprint_message += ' with fingerprint %s' % actual_hash
+        if ssl_cipher:
+            fingerprint_message += ' using cipher %s' % ssl_cipher
+        fingerprint_message += os.linesep
+
+        if self.app_options.get('fingerprint', False):
+            self.log.info(fingerprint_message)
+        else:
+            self.log.trace(fingerprint_message)
 
 
 #######################################
@@ -614,8 +637,11 @@ class IMAPSSLinitMixIn(object):
 
             # Ensure cert is for server we're connecting to
             if ssl and self.conf['ca_certs']:
-                ssl_match_hostname(self.conn.ssl().getpeercert(),
-                               self.conf['server'])
+                ssl_match_hostname(
+                    self.conn.ssl().getpeercert(),
+                    self.conf.get('ssl_cert_hostname', None) 
+                        or self.conf['server']
+                )
 
             if ssl_fingerprints:
                 if not actual_hash:
@@ -752,6 +778,7 @@ class RetrieverSkeleton(ConfigurableBase):
         self.msgsizes = {}
         self.oldmail = {}
         self.__delivered = {}
+        self.deleted = {}
         self.mailbox_selected = False
         
     def setup_received(self, sock):
@@ -764,7 +791,7 @@ class RetrieverSkeleton(ConfigurableBase):
             self.remoteaddr = '[%s]:%s' % serveraddr[:2]
         else:
             # Shouldn't happen
-            self.log.warn('unexpected peer address format %s', str(serveraddr))
+            self.log.warning('unexpected peer address format %s' % str(serveraddr))
             self.remoteaddr = str(serveraddr)
         self.received_from = '%s (%s)' % (self.conf['server'], 
                                           self.remoteaddr)
@@ -1089,10 +1116,28 @@ class POP3RetrieverBase(RetrieverSkeleton):
         self.mailboxes = (None, )
         # Handle password
         if self.conf.get('password', None) is None:
-            self.conf['password'] = get_password(
-                self, self.conf['username'], self.conf['server'],
-                self.received_with, self.log
-            )
+            if self.conf.get('password_command', None):
+                # Retrieve from an arbitrary external command
+                command = self.conf['password_command'][0]
+                args = self.conf['password_command'][1:]
+                (rc, stdout, stderr) = run_command(command, args)
+                if stderr:
+                    self.log.warning(
+                        'External password program "%s" wrote to stderr: %s'
+                        % (command, stderr)
+                    )
+                if rc:
+                    # program exited nonzero
+                    raise getmailOperationError(
+                        'External password program error (exited %d)' % rc
+                    )
+                else:
+                    self.conf['password'] = stdout
+            else:
+                self.conf['password'] = get_password(
+                    self, self.conf['username'], self.conf['server'],
+                    self.received_with, self.log
+                )
         RetrieverSkeleton.initialize(self, options)
         try:
             self._connect()
@@ -1162,7 +1207,8 @@ class MultidropPOP3RetrieverBase(POP3RetrieverBase):
         if self.envrecipname in NOT_ENVELOPE_RECIPIENT_HEADERS:
             raise getmailConfigurationError(
                 'the %s header field does not record the envelope '
-                'recipient address'
+                    'recipient address'
+                % self.envrecipname
             )
         self.envrecipnum = 0
         try:
@@ -1365,8 +1411,9 @@ class IMAPRetrieverBase(RetrieverSkeleton):
     def close_mailbox(self):
         # Close current mailbox so deleted mail is expunged.  One getmail
         # user had a buggy IMAP server that didn't do the automatic expunge,
-        # so we do it explicitly here.
-        self.conn.expunge()
+        # so we do it explicitly here if we've deleted any messages.
+        if self.deleted:
+            self.conn.expunge()
         self.conn.close()
         self.write_oldmailfile(self.mailbox_selected)
         # And clear some state
@@ -1485,7 +1532,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
                 )
             self.log.debug('deleting message "%s"' % uid + os.linesep)
             response = self._parse_imapuidcmdresponse(
-                'STORE', uid, 'FLAGS', '(\Deleted)'
+                'STORE', uid, 'FLAGS', '(\Deleted \Seen)'
             )
         except imaplib.IMAP4.error, o:
             raise getmailOperationError('IMAP error (%s)' % o)
@@ -1566,7 +1613,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             self.log.warning('Could not fetch google imap extensions: %s' % o)
             return {}
 
-        if not response:
+        if not response or not response[0]:
             return {}
             
         ext = re.search(
@@ -1611,10 +1658,28 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         # Handle password
         if (self.conf.get('password', None) is None
                 and not (HAVE_KERBEROS_GSS and self.conf['use_kerberos'])):
-            self.conf['password'] = get_password(
-                self, self.conf['username'], self.conf['server'], 
-                self.received_with, self.log
-            )
+            if self.conf['password_command']:
+                # Retrieve from an arbitrary external command
+                command = self.conf['password_command'][0]
+                args = self.conf['password_command'][1:]
+                (rc, stdout, stderr) = run_command(command, args)
+                if stderr:
+                    self.log.warning(
+                        'External password program "%s" wrote to stderr: %s'
+                        % (command, stderr)
+                    )
+                if rc:
+                    # program exited nonzero
+                    raise getmailOperationError(
+                        'External password program error (exited %d)' % rc
+                    )
+                else:
+                    self.conf['password'] = stdout
+            else:
+                self.conf['password'] = get_password(
+                    self, self.conf['username'], self.conf['server'], 
+                    self.received_with, self.log
+                )
             
         RetrieverSkeleton.initialize(self, options)
         try:
@@ -1664,7 +1729,7 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             (typ, dat) = self.conn.capability()
             if dat == [None]:
                 # No response, don't update the stored capabilities
-                self.log.warn('no post-login CAPABILITY response from server\n')
+                self.log.warning('no post-login CAPABILITY response from server\n')
             else:
                 self.conn.capabilities = tuple(dat[-1].upper().split())
 
@@ -1792,7 +1857,8 @@ class MultidropIMAPRetrieverBase(IMAPRetrieverBase):
         if self.envrecipname in NOT_ENVELOPE_RECIPIENT_HEADERS:
             raise getmailConfigurationError(
                 'the %s header field does not record the envelope recipient '
-                'address'
+                    'address'
+                % self.envrecipname
             )
         self.envrecipnum = 0
         try:
